from sage.misc.lazy_import import lazy_import

lazy_import('train_track.free_group', 'FreeGroup')

lazy_import('train_track.free_group_automorphism',['FreeGroupMorphism', 'FreeGroupAutomorphism', 'free_group_automorphisms'])

lazy_import('train_track.inverse_alphabet','AlphabetWithInverses')
